{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "(Data_structures)=\n",
    "# Data Structures\n",
    "``` {index} Data structures\n",
    "```\n",
    "\n",
    "*Data structures* allow us to speed up our algorithms often by many orders of magnitude. They arrange the data in a way that is suitable for the problem and usually simplify the workings of the algorithm. In this notebook, we will implement some of the most popular data structures in computer science:\n",
    "\n",
    "## Array\n",
    "``` {index} Array\n",
    "```\n",
    "Arrays are lists of objects. In some programming languages, arrays are objects of fixed length and have a type (e.g. list of 10 integers). In Python, are implemented using `list`s, which provide a little more flexibility by hosting multiple types and variable length. Arrays/lists support the following operation in complexity:\n",
    "\n",
    "* Set item (at index i smaller than the length): $O(1)$\n",
    "* Get item: $O(1)$\n",
    "* Deletion is no really supported as we assume that the array is of fixed length, for Python lists it is $O(n)$.\n",
    "\n",
    "Arrays are used in many programming problems where collections of data show up."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 98,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "2\n"
     ]
    }
   ],
   "source": [
    "arr = [1,2,3]\n",
    "# constant\n",
    "arr[2] = 4\n",
    "# constant\n",
    "print(arr[1])"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Stack\n",
    "```{index} Stack\n",
    "```\n",
    "A stack is a structure that follows that **LIFO** (last in, first out) rule. Elements can be `push`ed on top of the stack or `pop`ped of it. Those are the only two operations that this structure supports:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 99,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1\n",
      "ERROR: Stack is empty\n",
      "2\n",
      "[1, 3]\n"
     ]
    }
   ],
   "source": [
    "class Stack:\n",
    "    def __init__(self):\n",
    "        self.contents = []\n",
    "    def pop(self):\n",
    "        if len(self.contents):\n",
    "            temp = self.contents[-1]\n",
    "            self.contents = self.contents[:-1]\n",
    "            return temp\n",
    "        else:\n",
    "            print(\"ERROR: Stack is empty\")\n",
    "    def push(self, elem):\n",
    "        self.contents.append(elem)\n",
    "\n",
    "st = Stack()\n",
    "st.push(1)\n",
    "print(st.pop())\n",
    "st.pop()\n",
    "\n",
    "# Python lists also support these operations, push is called append!\n",
    "l = [1,2]\n",
    "print(l.pop())\n",
    "l.append(3)\n",
    "print(l)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Operations/complexity:\n",
    "* push: $O(1)$\n",
    "* pop: $O(1)$\n",
    "\n",
    "One of popular problems utilizing stacks is checking whether an expression has balanced brackets. Let us limit ourselves to expressions consisting of `[{,},(,)]`. We will push the opening brackets on the stack and pop them when the closing bracket is approached. Of course, the types of brackets must match:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 109,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "{}() True\n",
      "{( False\n",
      "({()}) True\n",
      ") False\n",
      " True\n"
     ]
    }
   ],
   "source": [
    "def checkIfValidBracketing(expr):\n",
    "    # our stack :)\n",
    "    st = []\n",
    "    for i in expr:\n",
    "        # not a bracket\n",
    "        if i not in [\"{\",\"}\",\"(\",\")\"]:\n",
    "            continue\n",
    "        # push the opening bracket on the stack\n",
    "        elif i in [\"{\",\"(\"]:\n",
    "            st.append(i)\n",
    "        # check when closing bracket approached    \n",
    "        elif i in [\"}\",\")\"]:\n",
    "            try:\n",
    "                temp = st.pop()\n",
    "                if (i == \"}\" and temp == \"{\") or (i == \")\" and temp == \"(\"):\n",
    "                    continue\n",
    "                # invalid bracketing    \n",
    "                else:\n",
    "                    return False\n",
    "            except IndexError:\n",
    "                return False\n",
    "    return not len(st)\n",
    "\n",
    "# Basic testing\n",
    "print(\"{}()\",checkIfValidBracketing(\"{}()\"))    \n",
    "print(\"{(\",checkIfValidBracketing(\"{(\")) \n",
    "print(\"({()})\",checkIfValidBracketing(\"({()})\")) \n",
    "print(\")\",checkIfValidBracketing(\")\")) \n",
    "print(\"\",checkIfValidBracketing(\"\")) "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The following algorithm has $O(n)$ time and space complexity."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Linked list\n",
    "```{index} Linked list\n",
    "```\n",
    "Linked lists are the simplest node-based structures. They are inherently recursive. A linked list consists of a `head` pointing to the `next` which itself is a linked list. The base case is an empty list:\n",
    "\n",
    "```{figure} algo_images/LinkedList.png\n",
    ":width: 60%\n",
    "```\n",
    "\n",
    "A simple implementation would be:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 110,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "True\n",
      "0->\n",
      "True\n",
      "0->1->\n",
      "True\n",
      "0->2->1->\n",
      "True\n",
      "0->2->\n",
      "True\n",
      "True\n",
      "2->\n",
      "False\n",
      "True\n",
      "False\n"
     ]
    }
   ],
   "source": [
    "class Node:\n",
    "    def __init__(self,val):\n",
    "        self.val = val\n",
    "        self.next = None\n",
    "        \n",
    "class LinkedList:\n",
    "    def __init__(self):\n",
    "        self.head = None  \n",
    "    # insert at a position index  \n",
    "    # false is index out of range, true otherwise\n",
    "    def insert(self,x, index):\n",
    "        assert index >= 0\n",
    "        newNode = Node(x)\n",
    "        if not index:\n",
    "            newNode.next = self.head\n",
    "            self.head = newNode\n",
    "            return True\n",
    "        else:\n",
    "            curr = self.head \n",
    "            for _ in range(index-1):\n",
    "                if curr is None:\n",
    "                    return False\n",
    "                curr = curr.next\n",
    "            if curr is None:\n",
    "                return False\n",
    "            newNode.next = curr.next\n",
    "            curr.next = newNode\n",
    "            return True\n",
    "            \n",
    "    # remove item at index position  \n",
    "    # return true if the element was actually removed, \n",
    "    # false otherwise\n",
    "    def remove(self,index):\n",
    "        assert index >= 0\n",
    "        # Removing head\n",
    "        if not index:\n",
    "            if self.head is not None:\n",
    "                self.head = self.head.next\n",
    "                return True\n",
    "            else:\n",
    "                return False\n",
    "        # index > 0\n",
    "        else:\n",
    "            curr = self.head\n",
    "            for _ in range(index-1):\n",
    "                if curr is None:\n",
    "                    return False\n",
    "                curr = curr.next\n",
    "            if curr is None:\n",
    "                return False\n",
    "            else:\n",
    "                if curr.next is None:\n",
    "                    return False\n",
    "                else:\n",
    "                    curr.next = curr.next.next\n",
    "                    return True\n",
    "                \n",
    "    # returns true if x in the list, false otherwise            \n",
    "    def search(self,x):\n",
    "        curr = self.head\n",
    "        while curr is not None:\n",
    "            if curr.val == x:\n",
    "                return True\n",
    "            curr = curr.next\n",
    "        return False\n",
    "    \n",
    "    # for testing and nice representation\n",
    "    def __str__(self):\n",
    "        res = \"\"\n",
    "        curr = self.head\n",
    "        while curr is not None:\n",
    "            res += str(curr.val) + \"->\"\n",
    "            curr = curr.next\n",
    "        return res\n",
    "        \n",
    "    \n",
    "ll = LinkedList()\n",
    "print(ll)\n",
    "print(ll.insert(0,0))\n",
    "print(ll)\n",
    "print(ll.insert(1,1))\n",
    "print(ll)\n",
    "print(ll.insert(2,1))\n",
    "print(ll)\n",
    "print(ll.remove(2))\n",
    "print(ll)\n",
    "print(ll.search(2))\n",
    "print(ll.remove(0))\n",
    "print(ll)\n",
    "print(ll.search(0))\n",
    "print(ll.remove(0))\n",
    "print(ll.remove(0))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "All of these operations (`insert`,`remove`,`search`) take $O(n)$ which is usually not optimal. There are numerous versions of linked lists e.g. doubly linked list which has nodes that point both to the previous and next nodes. Linked lists are often used to implement other data structures such as queues."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Queue\n",
    "```{index} Queue\n",
    "```\n",
    "Ques are similar to stacks but utilize the **FIFO** (first in, first out) rule. Basic ques support `enqueue` and `dequeue` operations. A linked list implementation would be:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 111,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\n",
      "1->2->3->\n",
      "1\n",
      "2->3->\n",
      "2\n",
      "3->\n",
      "3\n",
      "\n",
      "None\n"
     ]
    }
   ],
   "source": [
    "class Node:\n",
    "    def __init__(self, item):\n",
    "        self.item = item\n",
    "        self.next = None\n",
    "        self.previous = None\n",
    "\n",
    "class Queue:\n",
    "    def __init__(self):\n",
    "        self.length = 0\n",
    "        self.head = None\n",
    "        self.tail = None\n",
    "\n",
    "    def enqueue(self, x):\n",
    "        newNode = Node(x)\n",
    "        if self.head == None:\n",
    "            self.head = self.tail = newNode\n",
    "        else:\n",
    "            self.tail.next = newNode\n",
    "            newNode.previous = self.tail\n",
    "            self.tail = newNode\n",
    "        self.length += 1\n",
    "\n",
    "    def dequeue(self):\n",
    "        if not self.length:\n",
    "            return None\n",
    "        item = self.head.item\n",
    "        self.head = self.head.next \n",
    "        self.length -= 1\n",
    "        if self.length == 0:\n",
    "            self.tail = None\n",
    "        return item\n",
    "    \n",
    "    def __str__(self):\n",
    "        res = \"\"\n",
    "        curr = self.head\n",
    "        while curr is not None:\n",
    "            res += str(curr.item) + \"->\"\n",
    "            curr = curr.next\n",
    "        return res\n",
    "\n",
    "queue = Queue()\n",
    "print(queue)\n",
    "queue.enqueue(1)\n",
    "queue.enqueue(2)\n",
    "queue.enqueue(3)\n",
    "print(queue)\n",
    "print(queue.dequeue())\n",
    "print(queue)\n",
    "print(queue.dequeue())\n",
    "print(queue)\n",
    "print(queue.dequeue())\n",
    "print(queue)\n",
    "print(queue.dequeue())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "With this implementation `enqueue` and `dequeue` are $O(1)$. Queues are used in numerous problems, such as graph traversals (BFS). "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Binary search tree (BST)\n",
    "```{index} Binary search tree (BST)\n",
    "```\n",
    "*Binary tree* is another node-based structure also recursive in nature. The tree consists of a *root* which points to at most two children which are also binary trees. *Binary search trees* are a type of binary trees which have the property that the key of the left child is no greater than the root key. Right child's key is greater than the root key. An example binary search tree is as follows:\n",
    "\n",
    "```{figure} algo_images/BST.png\n",
    ":width: 80%\n",
    "```\n",
    "\n",
    "An example implementation for BST with integers is as follows:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 112,
   "metadata": {},
   "outputs": [],
   "source": [
    "# implemented as a set, will not contain duplicates\n",
    "class BST:\n",
    "    def __init__(self,val=None):\n",
    "        self.val = val\n",
    "        self.left = None\n",
    "        self.right = None\n",
    "        \n",
    "    # true if the insert changes the BST    \n",
    "    def insert(self,x):\n",
    "        if self.val is None:\n",
    "            self.val = x\n",
    "            return True\n",
    "        # already in the set\n",
    "        elif self.val == x:\n",
    "            return False\n",
    "        elif self.val > x:\n",
    "            if self.left is None:\n",
    "                self.left = BST(x)\n",
    "                return True\n",
    "            else:\n",
    "                return self.left.insert(x)\n",
    "        else:\n",
    "            if self.right is None:\n",
    "                self.right = BST(x)\n",
    "                return True\n",
    "            else:\n",
    "                return self.right.insert(x)\n",
    "            \n",
    "    # seach for x in BST\n",
    "    def search(self,x):\n",
    "        if self.val is None:\n",
    "            return False\n",
    "        elif self.val == x:\n",
    "            return True\n",
    "        elif self.val > x:\n",
    "            if self.left is None:\n",
    "                return False\n",
    "            else:\n",
    "                return self.left.search(x)\n",
    "        else:\n",
    "            if self.right is None:\n",
    "                return False\n",
    "            else:\n",
    "                return self.right.search(x)            "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Node removal is a bit more involved and will not be covered. All basic operations can be done in $O(\\log(n))$. BST can be used to speed up many problems that require repetitive searching of the data. Consider the following problem:\n",
    "\n",
    "**Numbers smaller than the given number in an array** You are given an array of integers. You should provide a data structure which should return the number of elements smaller than any chosen element in the array.\n",
    "\n",
    "We will solve the question by constructing a BST out of the array that will store the the number of nodes to the left of the node (its rank):"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 113,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3\n",
      "8\n",
      "-1\n"
     ]
    }
   ],
   "source": [
    "class newNode:\n",
    "    def __init__(self, data):\n",
    "        self.data = data \n",
    "        self.left = self.right = None\n",
    "        self.leftSize = 0\n",
    "        \n",
    "# Inserting a new Node. \n",
    "def insert(root, data):\n",
    "    if root is None: \n",
    "        return newNode(data) \n",
    " \n",
    "    # Updating size of left subtree. \n",
    "    if data <= root.data: \n",
    "        root.left = insert(root.left, data) \n",
    "        root.leftSize += 1\n",
    "    else:\n",
    "        root.right = insert(root.right, data)\n",
    "    return root\n",
    " \n",
    "# Function to get Rank of a Node x. \n",
    "def getRank(root, x):\n",
    "    if root.data == x:\n",
    "        return root.leftSize \n",
    "    if x < root.data: \n",
    "        if root.left is None: \n",
    "            return -1\n",
    "        else:\n",
    "            return getRank(root.left, x)\n",
    "    else: \n",
    "        if root.right is None: \n",
    "            return -1\n",
    "        else: \n",
    "            rightSize = getRank(root.right, x)\n",
    "            if rightSize == -1:\n",
    "                # x not found in right sub tree, i.e. not found in stream\n",
    "                return -1\n",
    "            else:\n",
    "                return root.leftSize + 1 + rightSize\n",
    "arr = [5, 1, 4, 4, 5, 9, 7, 13, 3] \n",
    "root = None\n",
    "for i in range(len(arr)):\n",
    "    root = insert(root, arr[i]) \n",
    "print(getRank(root, 4))\n",
    "print(getRank(root, 13))\n",
    "print(getRank(root, 8))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This algorithm will work in $O(\\log(n))$ time on average. However, of the array was initially sorted the formed tree will be degenerate (it will form a linked list). We will learn how to solve such cases in AVL trees."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Heap\n",
    "```{index} Heap\n",
    "```\n",
    "*Heaps* (especially *binary heaps*) are tree-like structures that have two properties:\n",
    "\n",
    "1) They are a complete tree (all levels of the tree are filled except possibly the last level).\n",
    "2) It is either a min-heap or a max-heap. In the first case, the root is smaller or equal to its children. The opposite is true for the max heap.\n",
    "\n",
    "Let us implement a min-heap storing integers. It will support taking the top element (which is the minimum) as well as inserting new elements. Interestingly, we will implement it using a Python list. It is an important fact to spot that a parent of an element at index `i` is at index `i//2`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 114,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 2, 2, 5, 3, 3, 4, 6, 6, 8, 7, 3, 8]\n",
      "[1, 2, 2, 3, 3, 3, 4, 5, 6, 6, 7, 8, 8]\n"
     ]
    }
   ],
   "source": [
    "# utility functions\n",
    "def swap(arr, x, y):\n",
    "    temp = arr[x]\n",
    "    arr[x] = arr[y]\n",
    "    arr[y] = temp\n",
    "    \n",
    "class MinHeap:\n",
    "    def __init__(self):\n",
    "        self.contents = []\n",
    "        self.size = 0\n",
    "        \n",
    "    # utility function    \n",
    "    def heapifyUp(self, i):\n",
    "        while i // 2 > 0:\n",
    "            if self.contents[i-1] < self.contents[i // 2 - 1]:\n",
    "                swap(self.contents,  i-1, i // 2 - 1)\n",
    "            i //= 2\n",
    "    \n",
    "    def insert(self,val):\n",
    "        self.size += 1\n",
    "        self.contents.append(val)\n",
    "        self.heapifyUp(self.size)\n",
    "        \n",
    "    def heapifyDown(self,i):\n",
    "        while (2 * i + 1) < self.size:\n",
    "            mc = self.minChild(i)\n",
    "            if self.contents[i] > self.contents[mc]:\n",
    "                swap(self.contents, i, mc)\n",
    "            i = mc\n",
    "            \n",
    "    def minChild(self, i):\n",
    "        if 2 * i + 2 >= self.size:\n",
    "            return 2 * i + 1\n",
    "        else:\n",
    "            if self.contents[2 * i + 1] < self.contents[2 * i + 2]:\n",
    "                return 2 * i + 1\n",
    "            else:\n",
    "                return 2 * i + 2\n",
    "        \n",
    "    def getMin(self):\n",
    "        assert self.size > 0\n",
    "        self.size -= 1\n",
    "        res = self.contents[0]\n",
    "        swap(self.contents,0, self.size)\n",
    "        self.contents.pop()\n",
    "        self.heapifyDown(0)\n",
    "        return res\n",
    "        \n",
    "#Heapsort -  a O(nlog(n)) sorting algorithm.\n",
    "\n",
    "arr = [1,3,2,6,8,3,4,5,6,7,2,3,8]\n",
    "heap = MinHeap()\n",
    "\n",
    "for j in range(len(arr)):\n",
    "    heap.insert(arr[j])\n",
    "\n",
    "print(heap.contents)   \n",
    "arr = []\n",
    "while heap.size:\n",
    "    arr.append(heap.getMin())\n",
    "print(arr)    "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Heaps have many, many uses. One of such is HeapSort - a $O(n\\log(n))$ sorting algorithm (see above). Heaps are also used in *priority queues* - queues which have the elements sorted by some *priority* such as objects sorted by their value. "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## AVL Tree\n",
    "```{index} AVL Tree\n",
    "```\n",
    "Consider the case we insert a sorted list into a regular BST. We would end up with the following structure:\n",
    "\n",
    "\n",
    "```{figure} algo_images/DegenerateTree.png\n",
    ":width: 40%\n",
    "```\n",
    "\n",
    "This tree is *degenerate*. Search operations are $O(n)$ which is too slow for many implementations. To counter the degeneration we can utilize *self-balancing binary trees* such as the AVL tree (named after  Adelson-Velsky and Landis). This type of tree holds the invariant that for each node the difference in heights between left and right subtree is at most 1. Let us see how it is achieved.\n",
    "\n",
    "### AVL insertion\n",
    "\n",
    "1) The first stage of insertion is the same as in BST, we recurse down the tree until we find an appropriate spot for the new value. \n",
    "\n",
    "2) Each node in the AVL tree has a *height* (the distance between the node and the farthest leaf node in the children). We check if the difference in heights of children exceeds 1 in absolute value. If so, the tree is imbalanced. Such violation can occur in four possible ways (see the diagram below). We will use node *rotations* to solve the problem:\n",
    "\n",
    "\n",
    "```{figure} algo_images/AVL.png\n",
    ":width: 80%\n",
    "```\n",
    "The implementation of the AVL tree is presented below, please take a while to digest it. The topic is not trivial. Also, have a look at the comparison between the BST and AVL in the degenerate case:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 115,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "BST time: 0.538590669631958\n",
      "AVL time: 0.014953136444091797\n",
      "Height of the AVL Tree: 10\n"
     ]
    }
   ],
   "source": [
    "import time\n",
    "# we will support only insertions and deletions\n",
    "class AVLNode:\n",
    "    # has to be initialised with a root value\n",
    "    def __init__(self,val):\n",
    "        self.val = val\n",
    "        self.left = None\n",
    "        self.right = None\n",
    "        self.height = 0\n",
    "\n",
    "    # searching is the same as in regural BST    \n",
    "    def search(self,val):\n",
    "        if val == self.val:\n",
    "            return True\n",
    "        elif val < self.val:\n",
    "            if self.left is not None:\n",
    "                return self.left.search(val)\n",
    "            else:\n",
    "                return False\n",
    "        else:\n",
    "            if self.right is not None:\n",
    "                return self.right.search(val)\n",
    "            else:\n",
    "                return False\n",
    "            \n",
    "    def insert(self, root, val):\n",
    "        # Normal BST insert\n",
    "        if not root:\n",
    "            return AVLNode(val)\n",
    "        elif root.val  == val:\n",
    "            return root\n",
    "        elif val < root.val:\n",
    "            root.left = self.insert(root.left,val)\n",
    "        else:\n",
    "            root.right = self.insert(root.right,val)\n",
    "           \n",
    "        # update the height of the root\n",
    "        root.height = 1 + max(getHeight(root.left), getHeight(root.right))\n",
    "        # get the balance factor\n",
    "        balance = getBalance(root)\n",
    "        # consider 4 cases\n",
    "        # LL\n",
    "        if balance > 1 and val < root.left.val:\n",
    "            return rotateRight(root)\n",
    "        # RR\n",
    "        if balance < -1 and val > root.right.val:\n",
    "            return rotateLeft(root)\n",
    "        # LR\n",
    "        if balance > 1 and val > root.left.val:\n",
    "            root.left = rotateLeft(root.left)\n",
    "            return rotateRight(root)\n",
    "        # RL\n",
    "        if balance < -1 and val < root.right:\n",
    "            root.right = rotateRight(root.right)\n",
    "            return rotateLeft(root)\n",
    "        return root\n",
    "        \n",
    "# UTILITY FUNCTIONS\n",
    "#_____________________________________________        \n",
    "# how is the tree unbalanced?\n",
    "def getBalance(node):\n",
    "    if not node:\n",
    "        return 0\n",
    "    else:\n",
    "        return getHeight(node.left) - getHeight(node.right)\n",
    "    \n",
    "# get height of a node\n",
    "def getHeight(node):\n",
    "    if not node:\n",
    "        return 0\n",
    "    else:\n",
    "        return node.height\n",
    "    \n",
    "def rotateRight(node):\n",
    "    # do the rewiring \n",
    "    z = node\n",
    "    y = z.left\n",
    "    T3 = y.right\n",
    "    \n",
    "    y.right = z\n",
    "    z.left = T3\n",
    "    \n",
    "    # update heights\n",
    "    z.height = 1 + max(getHeight(z.left), \n",
    "                       getHeight(z.right)) \n",
    "    y.height = 1 + max(getHeight(y.left), \n",
    "                       getHeight(y.right)) \n",
    "    return y\n",
    "\n",
    "def rotateLeft(node):\n",
    "    # do the rewiring \n",
    "    z = node\n",
    "    y = z.right\n",
    "    T2 = y.left\n",
    "    \n",
    "    y.left = z\n",
    "    z.right = T2\n",
    "    \n",
    "    # update heights\n",
    "    z.height = 1 + max(getHeight(z.left), \n",
    "                       getHeight(z.right)) \n",
    "    y.height = 1 + max(getHeight(y.left), \n",
    "                       getHeight(y.right))\n",
    "    return y\n",
    "#_____________________________________________  \n",
    "\n",
    "# Change the code below according to your preference \n",
    "arr = list(range(1000))   \n",
    "\n",
    "t = time.time()\n",
    "bst = BST(0)\n",
    "avl = AVLNode(0)\n",
    "\n",
    "for i in arr:\n",
    "    bst.insert(i)\n",
    "    bst.search(i)\n",
    "\n",
    "print(\"BST time:\",time.time() - t)\n",
    "t = time.time()\n",
    "\n",
    "for i in arr:\n",
    "    avl = avl.insert(avl, i)\n",
    "    avl.search(i)\n",
    "    \n",
    "print(\"AVL time:\", time.time() - t)\n",
    "print(\"Height of the AVL Tree: \" + str(avl.height))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Hash Table\n",
    "```{index} Hash Table\n",
    "```\n",
    "In simple terms, a hash table is identical to Python `dict` dictionaries. They allow us to retrieve objects (given their key) in $O(1)$ time *on average*. They also allow for constant time insertion and deletion. Wait, what? Does it mean that they are ultimately the fastest data structures? - as you can probably guess it is not that simple. It will become clear as we will implement a simple hash table:\n",
    "\n",
    "A hash table is similar to an array. It has some fixed capacity which can be changed (not in constant time). The elements are indexed by the key values. For the hash table to work, the elements need to be indexed by integer values. A special function, called the *hash function* transforms the key values to integers. The basic utilities needed by a hash table are shown in the class below:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 116,
   "metadata": {},
   "outputs": [],
   "source": [
    "class HashTable:\n",
    "    def __init__(self,size = 8):\n",
    "        pass\n",
    "    \n",
    "    #Insert a new key-value pair or overwrite an existing one\n",
    "    def insert(self,key, value):\n",
    "        pass\n",
    "    \n",
    "    # Overload the [] index access \n",
    "    def __setitem__(self,key,value):\n",
    "        return self.insert(key,value)\n",
    "    \n",
    "    # Retreive the value given the key\n",
    "    def get(self, key):\n",
    "        pass\n",
    "    \n",
    "   # Overload the [] index access \n",
    "    def __getitem__(self,key):\n",
    "        return self.get(key)\n",
    "    \n",
    "    # Computes the hash position\n",
    "    def _hash(self,key):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Hash function\n",
    "\n",
    "First, we will choose a hash function which will map keys to particular positions in the array. There are multiple ways of approaching this problem. Two simple ones include:\n",
    "\n",
    "* **The mod method** In this case, we take the remainder of dividing the key by the table-size:\n",
    "\n",
    "`h(key) = key % table_size`\n",
    "\n",
    "This is a very fast hash function. However, many keys can have the same hash which results in *collisions* (e.g. 64 % 8 and 8 % 8 are both 0). We discuss those below.\n",
    "\n",
    "* **The multiplication method** First we multiply the key by `0 < c < 1` and take the *fractional part* of `k * c` which is `k * c - floor(k * c)`. This method leads to fewer overlaps and therefore we will use it in our implementation. Some values of `c` work better statistically. `c ~ (sqrt (5) – 1) / 2 ` is likely to work well.\n",
    "\n",
    "`h(key) = floor(table_size * (key * c - floor(key * c)))`\n",
    "\n",
    "Both of these functions can lead to inefficient hash tables with operations with $O(n)$ worst-case time. There are more advanced methods which solve this problem, which you are free to research [here](https://en.wikipedia.org/wiki/List_of_hash_functions).\n",
    "\n",
    "### Overlaps\n",
    "\n",
    "When two different keys are mapped to the same hash value we are dealing with an overlap. The way our implementation will tackle this problem is by *chaining*. Each entry in the table will be a linked list-like object. In case of an overlap, we will append at the end of such a list. Of course, in the case when all elements *collide* we will one big linked-list object. The role of our hash function is to minimise the probability of this happening.\n",
    "\n",
    "### Doubling\n",
    "\n",
    "Let `n` denote the number of elements in the hash table. The *load factor* is defined as `n / table_size`. As our table becomes more filled more cells (on average) become linked lists. To counter this effect we will double the size of the table when the load factor reaches a certain value (in our case $1$). We will not support deletion therefore shrinking is not essential.\n",
    "\n",
    "With all this knowledge let us provide an implementation:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 117,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "4\n",
      "7\n",
      "4\n",
      "0\n",
      "6\n",
      "4\n",
      "3\n",
      "1\n",
      "11\n",
      "8\n",
      "7\n",
      "3\n",
      "17\n",
      "8\n",
      "5\n",
      "5\n",
      "[[], [], [('lemo', 3), ('emz', 5)], [], [('ps', 0)], [('sdens', 4), ('jbzni', 3)], [('jcpgq', 17)], [('zjb', 7)], [('w', 4), ('cbl', 8)], [], [], [], [('xsfj', 5)], [('po', 7), ('sazku', 4), ('udl', 1), ('vv', 8)], [('i', 11)], [('n', 6)]]\n"
     ]
    }
   ],
   "source": [
    "import math, random\n",
    "# We will support string keys\n",
    "# it could be easily translated to integer keys\n",
    "# the table will be a 2d array of tuples (key,value)\n",
    "class HashTable:\n",
    "    def __init__(self,size = 8):\n",
    "        self.contents = []\n",
    "        self.size = size\n",
    "        self.load = 0\n",
    "        for _ in range(self.size):\n",
    "            self.contents.append([])\n",
    "       \n",
    "    #Insert a new key-value pair or overwrite an existing one\n",
    "    def insert(self,key, value):\n",
    "        elem = (key, value)\n",
    "        # calculate the hash value\n",
    "        hashVal = self._hash(key)\n",
    "        # insert the element (chaining)\n",
    "        # check in elem already in\n",
    "        if elem not in self.contents[hashVal]:\n",
    "            self.contents[hashVal].append(elem)\n",
    "            # increase the load\n",
    "            self.load += 1\n",
    "        # doubling\n",
    "        if self.load / self.size > 1:\n",
    "            # double the size\n",
    "            self.size *= 2\n",
    "            # make new contents table\n",
    "            newContents = []\n",
    "            for _ in range(self.size):\n",
    "                newContents.append([])\n",
    "            # insert the old elements    \n",
    "            for i in self.contents:\n",
    "                for j in i:\n",
    "                    hashVal = self._hash(j[0])\n",
    "                    # insert the element (chaining)\n",
    "                    newContents[hashVal].append(j)\n",
    "            self.contents = newContents   \n",
    "    \n",
    "    # Overload the [] index access \n",
    "    def __setitem__(self,key,value):\n",
    "        return self.insert(key,value)\n",
    "    \n",
    "    # Retreive the value given the key\n",
    "    def get(self, key):\n",
    "        hashVal = self._hash(key)\n",
    "        for i in self.contents[hashVal]:\n",
    "            if i[0] == key:\n",
    "                return i[1]\n",
    "        print(\"Key has no value associated\")\n",
    "    \n",
    "   # Overload the [] index access \n",
    "    def __getitem__(self,key):\n",
    "        return self.get(key)\n",
    "    \n",
    "    # Computes the hash position\n",
    "    def _hash(self,key):\n",
    "        k = sum(ord(c) for c in key)\n",
    "        c = (math.sqrt(5) - 1) / 2\n",
    "        return min(math.floor(self.size * (k * c - math.floor(k * c))),self.size - 1)\n",
    "    \n",
    "# Testing out     \n",
    "ht = HashTable()\n",
    "for _ in range(16):\n",
    "    temp = ''\n",
    "    for i in range(random.randint(1,5)):\n",
    "        temp += chr(random.randint(ord('a'),ord('z')))\n",
    "    ht[temp] = random.randint(0,17)\n",
    "    print(ht[temp])\n",
    "print(ht.contents)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## References\n",
    "* Cormen, Leiserson, Rivest, Stein, \"Introduction to Algorithms\", Third Edition, MIT Press, 2009\n",
    "* GeeksforGeeks, [Rank Elements in an Integer Stream (BST)](https://www.geeksforgeeks.org/rank-element-stream/)\n",
    "* Pythonds, [Binary Heap Implementation](https://runestone.academy/runestone/books/published/pythonds/Trees/BinaryHeapImplementation.html)\n",
    "* GeeksforGeeks, [AVL Insertion](https://www.geeksforgeeks.org/avl-tree-set-1-insertion/)\n",
    "* Sebastian Raschka, Github, [Implementing a Simple Hash Table](https://github.com/rasbt/algorithms_in_ipython_notebooks/blob/master/ipython_nbs/data-structures/hashtable-1.ipynb)\n",
    "* GeeksforGeeks, [Implementation of Hashing with Chaining in Python](https://www.geeksforgeeks.org/implementation-of-hashing-with-chaining-in-python/)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.8"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
